/*
 *	Example of simple character device driver (schar)
 */
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/config.h>

#include <linux/module.h>

#if defined(CONFIG_SMP)
#define __SMP__
#endif


#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>	/* for timers */
#include <linux/fs.h>		/* file modes and device registration */
#include <linux/poll.h>		/* for poll */
#include <linux/wrapper.h>	/* mem_map_reserve,mem_map_unreserve */
#include <linux/proc_fs.h>
#include <linux/sysctl.h>
#include <linux/init.h>


#include <linux/netdevice.h>   // struct device and dev_xxx()
#include <linux/etherdevice.h> // eth_xxx()

#include <asm/io.h>


#include "schar.h"
/* settable parameters */
static char *schar_name = NULL;

/* forward declarations for _fops */
void get_event(void);
int cleanup_exit2(void);
void cleanup_module(void);
int netif_rx_hook(struct sk_buff *skb);
int init_module2(void);
int init_module(void);

static unsigned int schar_poll(struct file *file, poll_table *wait);
static ssize_t schar_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t schar_write(struct file *file, const char *buf, size_t count,loff_t *offset);
static int schar_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int schar_open(struct inode *inode, struct file *file);
static int schar_release(struct inode *inode, struct file *file);


// static struct wait_queue *schar_wq = NULL;
// static struct wait_queue *schar_poll_read = NULL;
static wait_queue_head_t schar_wq;
static wait_queue_head_t  schar_poll_read;


/* sysctl entries */
static char schar_proc_string[SCHAR_MAX_SYSCTL];
static struct ctl_table_header *schar_root_header = NULL;
static int schar_read_proc(ctl_table *ctl, int write, struct file *file,
			   void *buffer, size_t *lenp);

static ctl_table schar_sysctl_table[] = {
	{ DEV_SCHAR_ENTRY,	/* binary id */
	  "0",			/* name */
	  &schar_proc_string,	/* data */
	  SCHAR_MAX_SYSCTL,	/* max size of output */
	  0644,			/* mode */
	  0,			/* child - none */
	  &schar_read_proc },	/* set up text */
	{ 0 }
};

static ctl_table schar_dir[] = {
	{ DEV_SCHAR,		/* /proc/dev/schar */
	  "schar",		/* name */
	  NULL,
	  0,
	  0555,
	  schar_sysctl_table },	/* the child */
	{ 0 }
};

static ctl_table schar_root_dir[] = {
	{ CTL_DEV,		/* /proc/dev */
	  "dev",		/* name */
	  NULL,
	  0,
	  0555,
	  schar_dir },		/* the child */
	{ 0 }
};

/* module parameters and descriptions */
MODULE_PARM(schar_name, "s");
MODULE_PARM_DESC(schar_name, "Name of device");


MODULE_DESCRIPTION("schar, Sample character with ethernet hook");
MODULE_AUTHOR("S. Durkin");

/* durkin addition start */
static char *bufsend;
static int lsend;
static int levt;
#define MMT_BUF_SIZE 131072
#define PNT_RING_SIZE 100  
char *pnt_ring[PNT_RING_SIZE];
static int page_ring_r=0;
static int page_ring=0;
static char *buffer;
static char *buffer2;
static int nbuffer={0};
static unsigned short int nbuf={0};
static unsigned short int nbuf_r={0};
static char *buffer_r;
static char *buffer2_r;
static int nbuffer_r={0};
static int buferr={0};
static long int nbufleft={0};
static int long nbufdump={0};
static int long buffers={0};
static int endcond={0};
static int wakecond={1};
static int blocking={0};
static int long dumponoff={0};
static int long nbufl={0};
static int long nbufl_r={0};
static int dump={0};
static int inuse=0;


int init_module2(void)
{
  int i;
  for(i=0;i<PNT_RING_SIZE;i++){
        pnt_ring[i] =kmalloc(MMT_BUF_SIZE,GFP_KERNEL);
	if (!pnt_ring[i]) {
			printk(KERN_INFO "lsd: failed kmalloc in init_module2 \n");
        }}
  for(i=0;i<PNT_RING_SIZE;i++)*(unsigned short int *)pnt_ring[i]=0xffff;
  page_ring=0;
  buffer=pnt_ring[0];
  buffer2=pnt_ring[0];
  page_ring_r=0;
  buffer_r=pnt_ring[0];
  buffer2_r=pnt_ring[0];
  nbuf=0;
  nbuf_r=0;
  nbuffer=0;
  nbuffer_r=0;
  buferr=0;
  nbufleft=0;
  nbufdump=0;
  buffers=0;
  nbufl=0;
  nbufl_r=0;
  return 0;
}

int netif_rx_hook(struct sk_buff *skb)
{ char *pnt;
  char *pnt2;
  int i,icnt;
  static int ileft={0};
  // pnt2=skb->data-30;
  // for(i=0;i<40;i++)printk(KERN_INFO "%02x ",pnt2[i]&0xff);printk(KERN_INFO "\n");

  // printk(KERN_INFO "lsdtst2 %d %d %d \n",nbuf,nbuffer,skb->len);
  // printk(KERN_INFO "lsdput %d %d \n",page_ring,nbuffer);
  if(dumponoff==1){
    pnt2=skb->data-9;
    // printk(KERN_INFO " lsd hook: %d %02x %02x \n",skb->len,pnt2[0]&0xff,pnt2[8]&0xff);
    
    ileft=page_ring-page_ring_r;
    if(ileft<0)ileft=ileft+PNT_RING_SIZE;
    if((ileft>=PNT_RING_SIZE-2)&&((pnt2[0]&0xf0)==0x90)&&((pnt2[8]&0xf0)==0xa0))dump=1;    
    if((ileft<PNT_RING_SIZE-2)&&((pnt2[0]&0xf0)==0x90)&&((pnt2[8]&0xf0)==0xa0))dump=0;
    if((ileft>=PNT_RING_SIZE-2)&&((pnt2[0]&0xf0)==0x80)&&(skb->len==50))dump=1;    
    if((ileft<PNT_RING_SIZE-2)&&((pnt2[0]&0xf0)==0x80)&&(skb->len==50))dump=0;
    // printk(KERN_INFO "lsd hook: %d %d %d \n",PNT_RING_SIZE,ileft,dump);
    if(dump==1){
      nbufdump=nbufdump+1;
      goto PACKDUMP;
    }
  }
  inuse=1;
  pnt=buffer2;

  *(unsigned short int *)buffer2=0xffff;
  nbuffer=nbuffer+2;
  buffer2=buffer2+2;
  // jtag *(int *)buffer2=skb->len+14;
  *(int *)buffer2=skb->len+18;

  nbuffer=nbuffer+2;
  buffer2=buffer2+2;
 
  // jtag  icnt=(skb->len+12)>>2;
    icnt=(skb->len+16)>>2;

    for(i=0;i<icnt;i++){
      // jtag   *(unsigned long int *)(buffer2+i*4)=*(unsigned long int *)(skb->data+i*4-10);
       *(unsigned long int *)(buffer2+i*4)=*(unsigned long int *)(skb->data+i*4-14);

    }
        // jtag nbuffer=nbuffer+skb->len+10;
    // jtag buffer2=buffer2+skb->len+10;
    // jtag buffers=buffers+ skb->len+14;
   
   nbuffer=nbuffer+skb->len+14;
   buffer2=buffer2+skb->len+14;
   buffers=buffers+ skb->len+18;
  
   
   // for(i=0;i<5;i++)printk(KERN_INFO " fillbuf %d %02x %02x \n",i,pnt[2*i]&0xff,pnt[2*i+1]&0xff);
  if(nbuffer+9018> MMT_BUF_SIZE){
    // printk(KERN_INFO "lsdadd reset  %d %d %d %d \n",nbuffer+9000,MMT_BUF_SIZE,page_ring,PNT_RING_SIZE); 
  page_ring=page_ring+1;
  if(page_ring>=PNT_RING_SIZE)page_ring=0;

  buffer2=pnt_ring[page_ring];
  nbuffer=0;
  if((*(unsigned short int *)buffer2!=0xffff)&&(*(unsigned short int *)buffer2!=0xfffe)){
    // printk(KERN_INFO " buf err %04x \n",*(unsigned short int *)buffer2); 
    // printk(KERN_INFO "lsd hook: %d %d %d %d \n",PNT_RING_SIZE,ileft,dump,skb->len);

    buferr=1;
  }
  }
   *(unsigned short int *)buffer2=0xffff;
   nbuf=nbuf+1;
   if(nbuf==0xfffe)nbuf=0;
   *(unsigned short int *)pnt=nbuf;
   nbufl=nbufl+1;
   nbufleft=nbufleft+1;
   if(blocking==1){
   if(wakecond==0&&nbufleft>0){
                       wakecond=1;
		       wake_up_interruptible(&schar_wq);
		       wake_up_interruptible(&schar_poll_read); 
   }
   }
   inuse=0;
PACKDUMP:
  kfree_skb(skb);
  return 1;
}


/*

int netif_rx_hook(struct sk_buff *skb)
{
  kfree_skb(skb);
  return 1;
}

*/




/* ********************* schar driver taken from Linux Programming ***********
   2nd Edition, Richard Stones, Neil Mathews           */



static struct file_operations schar_fops = {
	read: schar_read,
        write: schar_write,
	poll: schar_poll,
	ioctl: schar_ioctl,
	open: schar_open,
	release: schar_release,
};

static int schar_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
  int i;
  int tdumponoff;

	/* make sure that the command is really one of schar's */
	if (_IOC_TYPE(cmd) != SCHAR_IOCTL_BASE)
		return -ENOTTY;
		
	switch (cmd) {

		case SCHAR_RESET: {
                 // start dumping 
		  tdumponoff=dumponoff;
                  dumponoff=1;
                  dump=1;
                  page_ring=PNT_RING_SIZE-2;
                  page_ring_r=0;
                  current->state=TASK_INTERRUPTIBLE;
                  if(inuse==1)schedule_timeout(20); 
                  // now do initializing
                  for(i=0;i<PNT_RING_SIZE;i++)*(unsigned short int *)pnt_ring[i]=0xffff;
                  page_ring=0;
                  buffer=pnt_ring[0];
                  buffer2=pnt_ring[0];
                  page_ring_r=0;
                  buffer_r=pnt_ring[0];
                  buffer2_r=pnt_ring[0];
                  nbuffer=0;
                  nbuffer_r=0;
                  nbuf=0;
                  nbuf_r=0;
                  buferr=0;
                  nbufleft=0;
                  nbufdump=0;
                  buffers=0;
                  nbufl=0;
                  nbufl_r=0;
		  // stop dumping    
		  dumponoff=tdumponoff;
                  dump=0; 
                  page_ring=0;
                  page_ring_r=0; 

		  // printk(KERN_INFO "ioct:l SCHAR_RESET \n");
		return 0;
		}

		case SCHAR_END: { 
                endcond=1;
                wakecond=1; 
                // printk(KERN_INFO "ioct:l SCHAR_END \n");
                if(blocking==1){
		wake_up_interruptible(&schar_wq);
		wake_up_interruptible(&schar_poll_read);
                } 
	        return 0;
		}

	        case SCHAR_BLOCKON: {
		blocking=1;
                endcond=0;
                printk(KERN_INFO "ioct:l BLOCKON \n");
                return 0;
                }

	        case SCHAR_BLOCKOFF: {
		blocking=0;
                endcond=0;
                // printk(KERN_INFO "ioct:l BLOCKOFF \n");
                return 0;
                }

	        case SCHAR_DUMPON: {
		  dumponoff=1;
                // printk(KERN_INFO "ioct:l DUMPON \n");
                return 0;
                }

	        case SCHAR_DUMPOFF: {
		  dumponoff=0;
                // printk(KERN_INFO "ioct:l DUMPOFF \n");
                return 0;
                }

		default: {
		  // MSG("ioctl: no such command\n");
			return -ENOTTY;
		}
	}

	/* to keep gcc happy */
	return 0;
}


static unsigned int schar_poll(struct file *file, poll_table * wait)
{
	unsigned int mask = 0;

	poll_wait(file, &schar_poll_read, wait);
	
	/* if the pool contains data, a read will succeed */
	if (nbufleft > 0)
		mask |= POLLIN | POLLRDNORM;
		
	/* a write always succeeds */
	mask |= POLLOUT | POLLWRNORM;
	
	return mask;
}


static int schar_read_proc(ctl_table *ctl, int write, struct file *file,
			   void *buffer, size_t *lenp)
{
	int len = 0;
	

	/* someone is writing data to us */
	if (write) {
		char *tmp = (char *) get_free_page(GFP_KERNEL);
	   
		if (tmp) {
			free_page((unsigned long)tmp);
			file->f_pos += *lenp;
		}
		return 0;
	}
	
	len += sprintf(schar_proc_string, "gigabit driver\n\n");
	len += sprintf(schar_proc_string+len, "written\t\t\t%ld\n",nbufl);
	len += sprintf(schar_proc_string+len, "read\t\t\t%ld\n",nbufl_r);
        len += sprintf(schar_proc_string+len, "nleft\t\t\t%ld\n",nbufleft);
        len += sprintf(schar_proc_string+len, "ndumped\t\t\t%ld\n",nbufdump);
        len += sprintf(schar_proc_string+len, "bytes \t\t\t%ld\n",buffers);
        len += sprintf(schar_proc_string+len, "page_ring \t\t\t%d\n",page_ring);
        len += sprintf(schar_proc_string+len, "page_ring_r \t\t\t%d\n",page_ring_r);

					
	
	*lenp = len;
	return proc_dostring(ctl, write, file, buffer, lenp);
	
	
}


/* increment and decrement usage count when someone is accessing
   /proc/sys/dev/schar */
static void schar_fill_inode(struct inode *inode, int fill)
{
	if (fill) {
		MOD_INC_USE_COUNT;
	} else {
		MOD_DEC_USE_COUNT;
	}
}

static ssize_t schar_read(struct file *file, char *buf, size_t count,
			  loff_t *offset)
{ unsigned short int tbuf[3];
  int tsend;
  unsigned short nbuf_rt;
 
 if(blocking==1&&nbufleft<=0&&endcond!=1){
                wakecond=0;
		interruptible_sleep_on_timeout(&schar_wq,100);
		if (signal_pending(current))return -EINTR;
  } 

  //  printk(KERN_INFO "schar_read: nbufleft %d \n",nbufleft);
  if(nbufleft>0){
        get_event();
        nbuf_rt=nbuf_r;
        nbuf_rt=nbuf_rt+1;
        if(nbuf_rt==0xfffe)nbuf_rt=0;
	// printk(KERN_INFO "schar_read: levt nbuf_r %d %d %ld %ld %d \n",levt,nbuf_r,nbufl_r,nbufl,nbufleft);
	// if((levt!=nbuf_rt)&&(levt!=0xffff)&&(levt!=0xfffe)&&(levt!=nbuf_r)){
       if((levt!=nbuf_rt)&&(levt!=0xffff)&&(levt!=0xfffe)){
	 // printk(KERN_INFO " lsderror 2 %d %d %d \n",nbuf_r,nbuf,levt);
	  tbuf[0]=2;
          tbuf[1]=levt;
          tbuf[2]=lsend;
          tsend=6; 
           if (copy_to_user(buf,tbuf,tsend))
		return -EFAULT;  
	   nbuf_r=levt;
           count=6;
           file->f_pos += count;
	   }	

        if(buferr==1){
          // printk(KERN_INFO "lsd error 4\n");
          buferr=2;
	  tbuf[0]=4;
          tbuf[1]=levt;
          tbuf[2]=lsend; 
	  tsend=6;
           if (copy_to_user(buf,tbuf,tsend))
		return -EFAULT;
           count=6;
	   file->f_pos += count;
	   return count;  
        }	

        if(levt==0xfffe||levt==0xffff){ 
	  // printk(KERN_INFO "lsd error 3\n");
	  tbuf[0]=3;
          tbuf[1]=levt;
          tbuf[2]=lsend; 
	  tsend=6;
           if (copy_to_user(buf,tbuf,tsend))
		return -EFAULT;  
           count=6;
	   file->f_pos += count;
	   return count;
        }
        if(lsend>9000){
          // printk(KERN_INFO "lsd error 1 lsend %d \n",lsend);
	  tbuf[0]=1;
          tbuf[1]=levt;
          tbuf[2]=lsend; 
          tsend=6; 
          if (copy_to_user(buf,tbuf,tsend))
		return -EFAULT;
          count=6;
	  file->f_pos += count;
	  return count;
        }
        *(unsigned short int *)(bufsend-4)=0xfffe;
        if (copy_to_user(buf,bufsend,lsend))
	  return -EFAULT;
        count=lsend;
	file->f_pos += count; 
        nbuf_r=nbuf_r+1;
        if(nbuf_r==0xfffe)nbuf_r=0;
        nbufl_r=nbufl_r+1;
        nbufleft=nbufl-nbufl_r;
	//  nbufleft=nbufleft-1;
	//  printk(KERN_INFO " %ld %ld %ld %ld \n",nbufleft,nbufl_r,nbufl,nbufleft+nbufl_r); 
	return count;
     }
  
     if(endcond==1){
          tbuf[0]=6;
          tbuf[1]=0;
          tbuf[2]=0;
      }else{
          tbuf[0]=3;
          tbuf[1]=levt;
          tbuf[2]=lsend;  
      }
     // printk(KERN_INFO " no data or end cond \n");
	  tsend=6;
           if (copy_to_user(buf,tbuf,tsend))
		return -EFAULT;
           count=6;
	   file->f_pos += count;
           endcond=0;
	   return count;
}



static int schar_open(struct inode *inode, struct file *file)
{
	/* increment usage count */
	MOD_INC_USE_COUNT;
	
	// printk(KERN_INFO "lsdmajor: %d minor: %d\n", MAJOR(inode->i_rdev), MINOR(inode->i_rdev));

	return 0;
}
static int schar_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}


int init_module(void)
{
	int res;
	
	if (schar_name == NULL)
		schar_name = "schar";
		
	
	/* register device with kernel */
	res = register_chrdev(SCHAR_MAJOR, schar_name, &schar_fops);
	if (res) {
		printk(KERN_INFO "can't register device with kernel\n");
		return res;
	}
	
	/* register proc entry */
	schar_root_header = register_sysctl_table(schar_root_dir, 0);
	//	schar_root_dir->child->de->fill_inode = &schar_fill_inode;
	
        init_waitqueue_head(&schar_wq);
        init_waitqueue_head(&schar_poll_read);

        init_module2();

	/* all done */	
	return 0;
}

void cleanup_module(void)
{
	/* unregister device and proc entry */
        cleanup_exit2();
	unregister_chrdev(SCHAR_MAJOR, "schar");
	if (schar_root_header)
		unregister_sysctl_table(schar_root_header);
      
	return;
}


void get_event(void)
{ // int j;  
  unsigned short int len;
  unsigned short int count2;
  //  int loop;
 // loop=0;
 // LOOPBACKR:
 // loop=loop+1;
 // for(j=0;j<100;j++);
 // if(loop>5000000)goto QUITLOOPR;
 // if(count2==0xffff)goto LOOPBACKR;
 // QUITLOOPR:
 count2=*(unsigned short int *)(buffer2_r);
 len=*(unsigned short int *)(buffer2_r+2);
 // printk(KERN_INFO "lsdget %d %d %d %d %d %d \n",count2,len,page_ring_r,nbuffer_r,page_ring,nbuffer);
 bufsend=buffer2_r+4;
 lsend=len-4;
 levt=count2;
 if(levt==0xffff)return;
 if(levt==0xfffe)return; 
 //   for(i=0;i<5;i++)printk(KERN_INFO "bufsend1 %d %02x %02x \n",i,bufsend[i*2]&0xff,bufsend[2*i+1]&0xff); 
  nbuffer_r=nbuffer_r+len;
 buffer2_r=buffer2_r+len;
 // if(nbuffer_r-3+9018>MMT_BUF_SIZE){
 if(nbuffer_r+9018>MMT_BUF_SIZE){
   page_ring_r=page_ring_r+1;
  if(page_ring_r>=PNT_RING_SIZE)page_ring_r=0;
  buffer2_r=pnt_ring[page_ring_r];
  nbuffer_r=0;
  } 
}

int cleanup_exit2(void)
{
  int i;
  //	printk(KERN_CRIT " lsd clean: cleanup memory \n");
  for(i=0;i<PNT_RING_SIZE;i++)kfree(pnt_ring[i]);
  return 0;
}

static ssize_t schar_write(struct file *file, const char *buf, size_t count,
			   loff_t *offset)
{
  //  struct sock *sk;
  // char *sbuf;
  // char *pnt2;
  // unsigned char *addr;
  int len;
  int err;
  // int i;
  static struct net_device *dev;
  static struct sk_buff *skb;
  static unsigned short proto=0;
  // ntk(KERN_INFO " length %d \n",count);

  // sbuf=kmalloc(9000,GFP_KERNEL);
  len=count;
  dev=dev_get_by_name("eth1");
  err=-ENODEV;
  if (dev == NULL)
   goto out_unlock;
            
/*
 *      You may not queue a frame bigger than the mtu. This is the lowest level
 *      raw protocol and you must do your own fragmentation at this level.
*/
                
  err = -EMSGSIZE;
  if(len>dev->mtu+dev->hard_header_len)
  goto out_unlock;
     
  err = -ENOBUFS;
  //  skb = sock_wmalloc(sk, len+dev->hard_header_len+15, 0, GFP_KERNEL);
  skb=dev_alloc_skb(len+dev->hard_header_len+15);   
/*
 *      If the write buffer is full, then tough. At this level the user gets to
 *      deal with the problem - do your own algorithmic backoffs. That's far
 *      more flexible.
*/
              
  if (skb == NULL) 
  goto out_unlock;
     
/*
*      Fill it in 
*/
              
/* FIXME: Save some space for broken drivers that write a
* hard header at transmission time by themselves. PPP is the
* notable one here. This should really be fixed at the driver level.
*/
   skb_reserve(skb,(dev->hard_header_len+15)&~15);
   skb->nh.raw = skb->data;
   proto=htons(ETH_P_ALL);
   /*     	if (dev->hard_header) {
		int res;
		err = -EINVAL;
                addr=NULL;
		res = dev->hard_header(skb, dev, ntohs(proto), addr, NULL, len);
			skb->tail = skb->data;
			skb->len = 0;
			} */
        			
                        skb->tail = skb->data;
			skb->len = 0;

/* Try to align data part correctly */
			/*      if (dev->hard_header) {
      skb->data -= dev->hard_header_len;
      skb->tail -= dev->hard_header_len;
      } */
 			
  
     //  printk(KERN_INFO " header length %d  \n",dev->hard_header_len);
/* Returns -EFAULT on error */
   //  err = memcpy_fromiovec(skb_put(skb,len), msg->msg_iov, len);
	
    err = copy_from_user(skb_put(skb,len),buf, count);
   // err = memcpy_fromio(skb_put(skb,len),sbuf,len);
   // printk(KERN_INFO " lsd: len count %d %d %02x  \n",len,count,*(skb->data+98)&0xff);
   skb->protocol = htons(ETH_P_ALL);
   skb->dev = dev;
   skb->priority = 0;
   // skb->pkt_type=PACKET_MR_PROMISC;
   skb->ip_summed=CHECKSUM_UNNECESSARY;
   if (err)
   goto out_free;
     
   err = -ENETDOWN;
   if (!(dev->flags & IFF_UP))
   goto out_free;
     
/*
*      Now send it
*/
     
   dev_queue_xmit(skb);
   dev_put(dev); 
   // printk(KERN_INFO " lsd: len count %d %d %02x  \n",len,count,*(skb->data+98)&0xff);

   // kfree(sbuf);
   return count;
     
   out_free:
     kfree_skb(skb);
   out_unlock:
     // if (dev)dev_put(dev);
      // kfree(sbuf);           
  return -EFAULT;
}












